import re
from .field import Field, ValidationError
import sys

class HandleParams:

    def __init__(
            self, params={}, sequence=None, ignored_fields=None, defaults=None, altered_fields=None
        ):
        self.sequence = list(sequence) if sequence else [] 
        self.unrecognized = []
        self.required_fields = set()
        self.altered_fields = dict(altered_fields) if altered_fields else {}
        self.defaults = dict(defaults) if defaults else {}
        for key in self.defaults.keys():
            if len(params.get(key,'') or '') > 0:
                self.defaults[key] = params[key]
        self.fields = {}
        self.ignored_fields = set(ignored_fields or [])

    def add_field(self, name, properties, **alter_cfg):
        
        properties['name'] = name

        variant = self.altered_fields.get(name)
        if variant is not None:
            if variant in alter_cfg:
                cfg = alter_cfg[variant]
                if not isinstance(cfg, dict):
                    raise RuntimeError(f'{name}: Not a dictionary for variant {variant}: {cfg}')
                alter_cfg = dict(cfg)
            else:
                # A variant is requested that is not known / supported.
                known_keys = ','.join(alter_cfg)
                raise RuntimeError(f"Field-def {name}: No variant {variant} (keys: {known_keys})")
        else:
            alter_cfg = {}
 
        for attr, acfg in alter_cfg.items() :
            if attr.startswith('wrap_'):
                attr = attr[5:]
                wrap = True
            else:
                wrap = False
            orig = properties.get(attr)
            properties[attr] = lambda orig, *args, **kwargs: acfg(orig, *args, **kwargs) if wrap else acfg
        if properties.pop('required', False) == True:
            self.required_fields.add(name)

        field = Field(**properties) 

        default = self.defaults.get(name)
        if default is not None:
            default = filter_value(field.apply_filters, name, default)
            try:
                if field.validate(default): field.value = default 
            except RuntimeError as e: # HandleParams.ValidationError raised spurious errors when
                if len(default) > 0:  # used flask is used with uwsgi
                    raise

        if hasattr(field, 'digest_key') or field.multiple:
            self.sequence.append(name)

        self.fields[name] = field

    def validate_params(self, params, set_invalid_value=True):
        success = True

        ignore = self.ignored_fields
        validate_seq = [*self.sequence]
        defer_validations = set(validate_seq)

        tmp = {}

        ## raise RuntimeError(type(params))

        for fname in self.required_fields:
            if len(params.get(fname, '')) == 0:
                params[fname] = None

        fieldlist = open('/tmp/validated_fields.txt','w')
        def paramsiter():
            for fldname, fldvalue in params.items():
                if isinstance(fldvalue, list):
                    for v in fldvalue:
                        yield fldname, v
                else:
                    yield fldname, fldvalue
        
        ## raise RuntimeError(list(paramsiter()))        
        ## raise RuntimeError(list(paramsiter()))
        ## raise RuntimeError(params.items())
        ## raise RuntimeError("defer_validations", list(defer_validations))
        ## if "foerderung" in defer_validations:
        ##     raise RuntimeError("foerderung in defer_valitations.")
        ## else:
        ##     raise RuntimeError("foerderung not in defer_validations.")
        ## for fldname, fldvalue in params.items():
        ##     raise RuntimeError(params.items())
        ##     if fldname == 'foerderung':
        ##         raise RuntimeError("type", type(fldvalue), fldvalue)
        ##     if fldname == 'foerderung' and len(fldvalue) > 1:
        ##         raise RuntimeError(fldvalue)
        for fldname, fldvalue in paramsiter():
            ## raise RuntimeError(fldname)
            ## if len(fldvalue) > 1:
            ##     raise RuntimeError(fldvalue)
            ## if fldname == 'foerderung':
            ##     raise RuntimeError(fldname + ' ' + fldvalue, type(fldvalue))
            if fldvalue is not None and len(fldvalue) == 0:
                continue
            if ( fldname not in defer_validations
                 and fldname in self.fields
               ):
                ## raise RuntimeError(fldname)
                ## if fldname == 'foerderung':
                ##     raise RuntimeError('test')
                ##     raise RuntimeError(fldname + ' ' + fldvalue)
                field = self.fields[fldname]
                value = filter_value(field.apply_filters, fldname, fldvalue)
                try:
                    this_success = field.validate(value)
                    ## von debugging sessions
                    if fldname == 'dsgvochecked' and this_success is False:
                        ## raise RuntimeError("this_success: " + this_success)
                        raise RuntimeError(value)

                except HandleParams.ValidationError as e:
                    field.error_msg = str(e)
                    field.is_valid = False
                    print(fldname, file=fieldlist)
                    this_success = False
                else:
                    field.is_valid = True
                take_this_value = value is not None if set_invalid_value else this_success
                if take_this_value:
                    field.value = value
                if not(success and this_success):
                    success = False

            elif not fldname in ignore:
                #if hasattr(self.fields[fldname], 'digest_key'):
                #    raise RuntimeError(fldname)
                #elif fldname.startswith("nachweise"):
                #    raise RuntimeError(fldname)
                ## raise RuntimeError("validate_seq: ", validate_seq)
                for fname in validate_seq: 
                    key_tmpl = {}
                    field = self.fields[fname]

                    if hasattr(field, 'digest_key'):
                        key_tmpl = field.digest_key(fldname)
                        if not key_tmpl: continue
                        value = filter_value(field.apply_filters, fname, fldvalue)
                        key_tmpl['value'] = value
                        value = key_tmpl
                        tmp.setdefault(fname, {})
                        slot = tmp[fname]

                    elif fldname == fname:
                        value = filter_value(field.apply_filters, fname, fldvalue)
                        slot = tmp
                        ## raise RuntimeError("slot: ",slot)
                    else:
                        continue

                    if field.multiple:
                        slot.setdefault(fldname, [])
                        slot[fname].append(value)
                        ## raise RuntimeError(fldname + ' ' + value)
                    else:
                        slot[fldname] = value

        ## raise RuntimeError(tmp) 
        fullfilled = False
        for fldname, fldvalue in tmp.items():
            field = self.fields[fldname]
            try:
                ##raise RuntimeError("fldvalue2: ", fldvalue)
                this_success = field.validate(fldvalue)
            except HandleParams.ValidationError as e:
                field.error_msg = str(e)
                field.is_valid = False
                this_success = False
            else:
                field.is_valid = True
            take_this_value = fldvalue is not None if set_invalid_value else this_success
            if take_this_value:
                field.value = fldvalue
                fullfilled = True
            if not(success and this_success):
                success = False
        #if not fullfilled:
        #    self.unrecognized.append(fldname)
        #    continue
    ##raise RuntimeError(tmp)
        return success

ADD_FILTERS = {}

def filter_value(filters_to_apply, name, value):
    if isinstance(value, str):
        value = re.sub(r'[\;\"\(\)\{\}\\\`\|\%\r\<\>\&\0]', '', value)
        for fname in filters_to_apply:
            if not fname in ADD_FILTERS:
                raise RuntimeError(f"No filter {fname}")
            if not isinstance(value, list):
                value = [list]
            for v in value:
                try:
                    value = FILTERS[fname](v)
                except HandleParams.ValidationError as e:
                    field.error_msg = str(e)
                    return
        return value
